# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS app
"""

import json
import os

from rattail import app as base
from rattail.config import make_config

import flet as ft

import wuttapos
from wuttapos.util import get_pos_batch_handler


class WuttaAppHandler(base.AppHandler):
    """
    Custom app handler for WuttaPOS
    """

    def get_title(self, **kwargs):
        kwargs.setdefault('default', 'WuttaPOS')
        return super().get_title(**kwargs)

    def get_node_title(self, **kwargs):
        kwargs.setdefault('default', 'WuttaPOS')
        return super().get_node_title(**kwargs)


def main(page: ft.Page):
    config = make_config()
    app = config.get_app()
    model = app.model

    page.title = f"WuttaPOS v{wuttapos.__version__}"
    page.window_full_screen = True
    # page.vertical_alignment = ft.MainAxisAlignment.CENTER

    # global defaults for button/text styles etc.
    page.data = {
        'default_button_height_pos': 100,
        'default_button_height_dlg': 80,
    }

    # nb. track current user, txn etc.
    page.shared = {}

    # TODO: this may be too hacky but is useful for now/dev
    if not config.production():
        path = os.path.join(config.appdir(), '.wuttapos.cache')
        if os.path.exists(path):
            with open(path, 'rt') as f:
                page.shared = json.loads(f.read())
        page.shared.pop('txn_display', None) # TODO
        page.shared.pop('cust_uuid', None) # TODO
        page.shared.pop('cust_display', None) # TODO
        if page.shared and page.shared.get('user_uuid'):
            handler = get_pos_batch_handler(config)
            session = app.make_session()
            user = session.get(model.User, page.shared['user_uuid'])
            if user:
                # TODO: should also filter this by terminal
                batch = handler.get_current_batch(user, create=False)
                if batch:
                    page.shared['txn_display'] = batch.id_str
                    page.shared['cust_uuid'] = batch.customer_uuid
                    if batch.customer:
                        key = app.get_customer_key_field()
                        value = getattr(batch.customer, key)
                        page.shared['cust_display'] = str(value or '') or None
                    else:
                        page.shared['cust_display'] = None
                else:
                    page.shared['txn_display'] = None
                    page.shared['cust_uuid'] = None
                    page.shared['cust_display'] = None
            else:
                page.shared['user_uuid'] = None
                page.shared['user_display'] = None
                page.shared['txn_display'] = None
                page.shared['cust_uuid'] = None
                page.shared['cust_display'] = None
            session.close()
        else:
            page.shared['user_display'] = None

    def clean_exit():
        if not config.production():
            # TODO: this may be too hacky but is useful for now/dev
            path = os.path.join(config.appdir(), '.wuttapos.cache')
            with open(path, 'wt') as f:
                f.write(json.dumps(page.shared))
        page.window_destroy()

    def keyboard(e):
        # exit on ctrl+Q
        if e.ctrl and e.key == 'Q':
            if not e.shift and not e.alt and not e.meta:
                clean_exit()

    page.on_keyboard_event = keyboard

    def window_event(e):
        if e.data == 'close':
            clean_exit()

    # cf. https://flet.dev/docs/controls/page/#window_destroy
    page.window_prevent_close = True
    page.on_window_event = window_event

    # TODO: probably these should be auto-loaded from spec
    from wuttapos.views.pos import POSView
    from wuttapos.views.login import LoginView

    # cf .https://flet.dev/docs/guides/python/navigation-and-routing#building-views-on-route-change

    def route_change(route):
        page.views.clear()
        if page.route == '/pos':
            page.views.append(POSView(config, '/pos'))
        elif page.route == '/login':
            page.views.append(LoginView(config, '/login'))
        page.update()

    # TODO: this was in example docs but not sure what it's for?
    # def view_pop(view):
    #     page.views.pop()
    #     top_view = page.views[-1]
    #     page.go(top_view.route)

    page.on_route_change = route_change
    # page.on_view_pop = view_pop

    # TODO: this may be too hacky but is useful for now/dev
    if not config.production() and page.shared.get('user_uuid'):
        page.go('/pos')
    else:
        page.go('/login')


# TODO: can we inject config to the main() via ft.app() kwargs somehow?
# pretty sure the `wuttapos open` command is trying to anyway..
def run_app(config=None):
    ft.app(target=main)


if __name__ == '__main__':
    run_app()
